# From https://stackoverflow.com/a/3041990, adapted for use in this script
def input_yn(question, default=True):
    valid = {"yes": True, "y": True, "ye": True,
             "no": False, "n": False}
    if default is None:
        prompt = " [y/n] "
    elif default:
        prompt = " [Y/n] "
    else:
        prompt = " [y/N] "

    while True:
        print(question + prompt, end='')
        choice = input().lower()
        if default is not None and choice == '':
            return default
        elif choice in valid:
            return valid[choice]
        else:
            print("Please respond with 'yes' or 'no' (or 'y' or 'n').")



SENTRY_SDK_URL = "https://0da75bab4671455ea1b7580cb93649f5@o444286.ingest.sentry.io/5645833"

FIRST_TIME_RUN_MSG = """
It looks like this is your first time running shippy.
We need to configure shippy before you can use it.

"""

BETA_CHUNK_UPLOAD_PROMPT_MSG = """
[BETA] shippy has a new upload method that can improve upload performance and reliability.
Would you like to try it out? Answer no if you would like to use the traditional method of uploading.
"""

NO_MATCHING_FILES_FOUND_ERROR_MSG = """
No files matching the submission criteria were detected.

Please check the following:
  1. Make sure you are in the correct directory.
  2. Please do not rename the build artifacts.

If you believe this is a problem with shippy, please contact maintainer support.
"""

UNHANDLED_EXCEPTION_MSG = """
shippy crashed for an unknown reason. :(
To figure out what went wrong, please pass along the full output.
----
URL of request: {}
Request response code: {}
Request response: {}
---
"""

# Defaults
DEFAULT_SHIPPY_CHUNKED_UPLOAD = "true"
DEFAULT_SHIPPY_CHUNKED_UPLOAD_SIZE = 10_000_000  # 10 MB



import os.path

import requests
from clint.textui.progress import Bar as ProgressBar
from requests_toolbelt import MultipartEncoder, MultipartEncoderMonitor

from .config import get_config_value
from .exceptions import LoginException, UploadException
from .constants import UNHANDLED_EXCEPTION_MSG

DEBUG = os.environ.get("SHIPPY_DEBUG", default=0)


def undef_response_exp(r):
    raise Exception(UNHANDLED_EXCEPTION_MSG.format(r.url, r.status_code, r.json()))


def get_server_version(server_url):
    """ Gets server version in semver format """
    version_url = "{}/maintainers/api/system/".format(server_url)
    r = requests.get(version_url)
    if r.status_code == 200:
        return r.json()['version']
    else:
        print("Warning: failed to retrieve server version information!")
        return None


def login_to_server(username, password, server_url):
    """ Logs in to server and returns authorization token """
    login_url = "{}/maintainers/api/login/".format(server_url)
    r = requests.post(login_url, data={'username': username, 'password': password})

    if r.status_code == 200:
        data = r.json()
        return data['token']
    elif r.status_code == 400:
        if r.json()['error'] == "blank_username_or_password":
            raise LoginException("Username or password must not be blank.")
    elif r.status_code == 404:
        if r.json()['error'] == "invalid_credential":
            raise LoginException("Invalid credentials!")
    else:
        undef_response_exp(r)


def upload_to_server(build_file, checksum_file, server_url, token, use_chunked_upload=False):
    print("Uploading build {}...".format(build_file))

    if use_chunked_upload:
        print("Using chunked upload method to upload...")
        chunked_upload(server_url, build_file, checksum_file, token)
    else:
        print("Using direct upload method to upload...")
        direct_upload(server_url, build_file, checksum_file, token)


def chunked_upload(server_url, build_file, checksum_file, token):
    device_upload_url = "{}/maintainers/api/chunked_upload/".format(server_url)

    chunk_size = int(get_config_value("shippy", "chunked_upload_size"))
    current_index = 0
    total_file_size = os.path.getsize(build_file)

    bar = ProgressBar(expected_size=total_file_size, filled_char='=')

    with open(build_file, 'rb') as build_file_raw:
        while chunk_data := build_file_raw.read(chunk_size):
            r = requests.put(device_upload_url, headers={
                "Authorization": "Token {}".format(token),
                "Content-Range": "bytes {}-{}/{}".format(current_index, current_index + len(chunk_data) - 1,
                                                         total_file_size),
            }, data={"filename": build_file}, files={'file': chunk_data})

            if r.status_code == 200:
                device_upload_url = "{}/maintainers/api/chunked_upload/{}/".format(server_url, r.json()['id'])
                current_index += len(chunk_data)
                bar.show(current_index)
            elif r.status_code == 429:
                print("shippy has been rate-limited.")
                import re
                wait_rate_limit(int(re.findall("\d+", r.json()['detail'])[0]))
            else:
                if DEBUG:
                    print("Status code received from server: {}".format(r.status_code))
                    with open('output.html', 'wb') as error_output_raw:
                        error_output_raw.write(r.content)
                raise UploadException("Something went wrong during the upload.")

    print("")  # Clear progress bar from screen

    # Finalize upload to begin processing
    r = requests.post(device_upload_url, headers={"Authorization": "Token {}".format(token)},
                      data={'md5': get_md5_from_file(checksum_file)})

    upload_exception_check(r, build_file)


def get_md5_from_file(checksum_file):
    with open(checksum_file, 'r') as checksum_file_raw:
        line = checksum_file_raw.readline()
        values = line.split(" ")
        return values[0]


def wait_rate_limit(seconds):
    import time
    while seconds:
        print("Will resume in {} seconds...".format(seconds), end='\r')
        time.sleep(1)
        seconds -= 1
    print(end='\x1b[2K\r')


def direct_upload(server_url, build_file, checksum_file, token):
    device_upload_url = "{}/maintainers/api/upload/".format(server_url)

    e = MultipartEncoder(fields={
        'build_file': (build_file, open(build_file, 'rb'), 'text/plain'),
        'checksum_file': (checksum_file, open(checksum_file, 'rb'), 'text/plain'),
    })

    bar = ProgressBar(expected_size=e.len, filled_char='=')

    def callback(monitor):
        bar.show(monitor.bytes_read)

    m = MultipartEncoderMonitor(e, callback)

    r = requests.post(device_upload_url, headers={
        "Authorization": "Token {}".format(token),
        "Content-Type": e.content_type
    }, data=m)

    upload_exception_check(r, build_file)


def upload_exception_check(r, build_file):
    if DEBUG:
        print("Received code: {}".format(r.status_code))
        print(r.json())

    if r.status_code == 200:
        print("Successfully uploaded the build {}!".format(build_file))
        return
    elif r.status_code == 400:
        if r.json()['error'] == "duplicate_build":
            raise UploadException("This build already exists in the system!")
        elif r.json()['error'] == "missing_files":
            raise UploadException("One of the required fields are missing!")
        elif r.json()['error'] == "file_name_mismatch":
            raise UploadException("The build file name does not match the checksum file name!")
        elif r.json()['error'] == "invalid_file_name":
            raise UploadException("The file name was malformed!")
        elif r.json()['error'] == "not_official":
            raise UploadException("The build is not official!")
        elif r.json()['error'] == "codename_mismatch":
            raise UploadException("The codename does not match the build file name!")
    elif r.status_code == 401:
        if r.json()['error'] == "insufficient_permissions":
            raise UploadException("You are not allowed to upload for this device!")
    elif r.status_code == 500:
        raise UploadException("Something went wrong with the server. Please contact the admins.")

    undef_response_exp(r)



import argparse
import glob
import hashlib
import os.path

import requests
import semver
import sentry_sdk

from .client import login_to_server, upload_to_server, get_server_version, get_md5_from_file
from .config import get_config_value, set_config_value
from .constants import *
from .exceptions import LoginException, UploadException
from .helper import input_yn
from .version import __version__, server_compat_version

ignore_errors = [KeyboardInterrupt]

sentry_sdk.init(SENTRY_SDK_URL, traces_sample_rate=1.0, release=__version__, ignore_errors=ignore_errors)


def main():
    print("Welcome to shippy (v.{})!".format(__version__))

    parser = argparse.ArgumentParser(description="Client-side tool for interfacing with shipper.")
    parser.add_argument('-c', '--chunk-size', action='store_true', help='Edit chunk size')
    args = parser.parse_args()

    if args.chunk_size:
        edit_chunk_size()
        return

    # Check for updates
    check_shippy_update()

    try:
        server_url = get_config_value("shippy", "server")
        token = get_config_value("shipper", "token")
    except KeyError:
        print(FIRST_TIME_RUN_MSG)
        server_url = get_server_url()
        token = get_token(server_url)

    check_server_compat(server_url)

    chunked_upload = (get_config_value("shippy", "chunked_upload").lower() == "true")

    # Search current directory for files
    builds = get_builds_in_current_dir()

    if len(builds) == 0:
        print(NO_MATCHING_FILES_FOUND_ERROR_MSG)
    else:
        print("Detected {} build(s):".format(len(builds)))
        for build in builds:
            print("\t{}".format(build))

        if len(builds) > 1:
            if not input_yn("Warning: you seem to be uploading multiple builds. Are you sure you want to continue?",
                            default=False):
                return

        for build in builds:
            # Check build file validity
            if not check_build(build):
                print("Invalid build. Skipping...")
                continue

            if input_yn("Uploading build {}. Start?".format(build)):
                try:
                    upload_to_server(build, "{}.md5".format(build), server_url, token,
                                     use_chunked_upload=chunked_upload)
                except UploadException as exception:
                    print(exception)


def check_server_compat(server_url):
    print("shippy is contacting the remote server... Please wait.")
    server_version = get_server_version(server_url)
    if semver.compare(server_version, server_compat_version) == -1:
        print("Warning: the server you're connecting to is out-of-date. shippy may not work properly.")
        print("If you know the server admin, please ask them to upgrade the server.")
        print(" * Reported server version: \t{}".format(server_version))
        print(" * Compatible version: \t\t{}".format(server_compat_version))
        if not input_yn("Are you sure you want to continue? Only proceed if you know what you are doing!",
                        default=False):
            exit(0)
    else:
        print("Finished compatibility check. No problems found.")


def check_shippy_update():
    print("Checking for updates...")
    r = requests.get("https://api.github.com/repos/ericswpark/shippy/releases/latest")
    latest_version = r.json()['name']

    if semver.compare(__version__, latest_version) == -1:
        print("Warning: shippy is out-of-date. We recommend updating with the following command:")
        print("\tpip3 install --upgrade shipper-shippy")
    else:
        print("Finished update check. shippy is up-to-date!")


def get_builds_in_current_dir():
    builds = []
    glob_match = 'Bliss-v*.zip'

    print("Detecting builds in current directory...")

    for file in glob.glob(glob_match):
        builds.append(file)

    return builds


def check_build(filename):
    """ Makes sure the build is valid """
    # Validate that there is a matching checksum file
    if not os.path.isfile("{}.md5".format(filename)):
        print("This build does not have a matching checksum file. ", end='')
        return False

    # Validate checksum
    md5_hash = hashlib.md5()
    with open(filename, "rb") as build_file:
        content = build_file.read()
        md5_hash.update(content)
    md5_hash = md5_hash.hexdigest()
    actual_hash = get_md5_from_file("{}.md5".format(filename))
    if md5_hash != actual_hash:
        print("This build's checksum is invalid. ", end='')
        return False

    build_slug, _ = os.path.splitext(filename)
    _, _, _, build_type, build_variant, _ = build_slug.split('-')

    # Check build type
    if build_type != "OFFICIAL":
        print("This build is not official. ", end='')
        return False

    # Check build variant
    valid_variants = ['gapps', 'vanilla', 'foss', 'goapps']
    if build_variant not in valid_variants:
        print("This build has an unknown variant. ", end='')
        return False

    return True


def get_server_url():
    server_url = input("Enter the server URL: ")

    while True:
        if "http" not in server_url:
            # noinspection HttpUrlsUsage
            print("Server URL seems to be missing the schema. Please add http:// or https:// to the server URL.")
        else:
            break
        server_url = input("Enter the server URL: ")

    if server_url[-1] == '/':
        server_url = server_url[:-1]

    set_config_value("shippy", "server", server_url)

    return server_url


def get_token(server_url):
    while True:
        from getpass import getpass

        username = input("Enter your username: ")
        password = getpass(prompt="Enter your password: ")

        try:
            token = login_to_server(username, password, server_url)
            set_config_value("shipper", "token", token)
            return token
        except LoginException:
            print("An error occurred logging into the server. Please try again.")


def edit_chunk_size():
    print("Now editing the chunk size.")
    user_size = input("Specify the chunk size in megabytes: ")
    set_config_value("shippy", "chunked_upload_size", user_size * 1_000_000)


if __name__ == "__main__":
    main()



import configparser

from pathlib import Path

from shippy.constants import DEFAULT_SHIPPY_CHUNKED_UPLOAD_SIZE, DEFAULT_SHIPPY_CHUNKED_UPLOAD

home_dir = str(Path.home())

# Constants
CONFIGURATION_FILE = "{}/.shippy.ini".format(home_dir)

# Load configuration
config = configparser.ConfigParser()
config.read(CONFIGURATION_FILE)


def get_config_value(section, key):
    try:
        value = config[section][key]
    except KeyError:
        # Set defaults and return
        value = get_default(section, key)
        if value:
            set_config_value(section, key, value)
        else:
            raise KeyError

    return value


def get_default(section, key):
    if section == "shippy":
        if key == "chunked_upload":
            return DEFAULT_SHIPPY_CHUNKED_UPLOAD
        elif key == "chunked_upload_size":
            return str(DEFAULT_SHIPPY_CHUNKED_UPLOAD_SIZE)


def set_config_value(section, key, value):
    config_init()
    config[section][key] = value
    config_save()


def config_init():
    if not config.has_section('shippy'):
        config.add_section('shippy')
    if not config.has_section('shipper'):
        config.add_section('shipper')


def config_save():
    with open(CONFIGURATION_FILE, 'w+') as config_file:
        config.write(config_file)
